/* AP26-boinc.c -- Geoffrey Reynolds, 11 September 2008.

   This file is a replacement for AP26-64.c in Jaroslaw Wroblewski's AP26
   sample implementation.

   See http://www.geocities.com/g_w_reynolds/AP26/AP26-boinc-src.zip for the
   latest version of this file.

   See http://www.math.uni.wroc.pl/~jwr/AP26/AP26.zip for the sample
   implementation containing the rest of the source required.

   See http://www.math.uni.wroc.pl/~jwr/AP26/faster32.zip for a patch to
   dramatically improve performance on 32-bit machines.

   See http://www.math.uni.wroc.pl/~jwr/AP26/AP26v3.pdf for information
   about how the algorithm works and for the copyleft notice.


   Compiling the source:
   ---------------------

   Unpack this archive (AP26-boinc-src.zip) along with the sample
   implementation source (AP26.zip) and patches (faster32.zip, if32.zip)
   from the same directory.
 
   To compile a standalone app for manual testing:

    gcc -O3 -o AP26 AP26-boinc.c

   To compile a BOINC app:

    gcc -O3 -DBOINC -o AP26 AP26-boinc.c -lstdc++ -lpthread -lboinc_api -lboinc

   For non-x86/x86_64/ppc64 architectures, add -lgmp to the gcc command line.


   Testing the executable:
   -----------------------

   To briefly test the AP26 executable, check that the file TEST-366384.txt
   matches the results file SOL-AP26.txt produced by executing:

     ./AP26 366384 366384 0


   Program operation:
   ------------------

   If search parameters are not given on the command line as

     AP26 [KMIN KMAX SHIFT [LINEAR]]

   then they must be present in a file AP26-ini.txt, separated by whitespace.

   If LINEAR is nonzero then KMIN, KMAX will first be set to f(KMIN), f(KMAX)
   respectively, where f(n) is the n'th positive integer not divisible by
   one of {29,31,37,41,43,47,53,59}.

   The search will begin at K=KMIN ITER=0 unless a file AP26-state.txt
   exists containing a checkpoint of the form

     KMIN KMAX SHIFT K ITER FREQ

   with KMIN KMAX SHIFT matching the initial search parameters, in which
   case the search will resume from that checkpoint.

   The search will continue up to and including K=KMAX. On completion
   AP26-state.txt will contain a checkpoint of the form

     KMIN KMAX SHIFT KMAX+1 0 FREQ

   No modification will be made to AP26-ini.txt.
   Periodic checkpoints will be written to AP26-state.txt.
   All search results will be appended to SOL-AP26.txt.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
// SSE2
#include <emmintrin.h>

#if defined(_MSC_VER) && (defined(_M_IX86) || defined(_M_X64))
# ifndef NO_LONG_DOUBLE
#  define NO_LONG_DOUBLE 1
# endif
#endif

#ifdef _MSC_VER
  typedef __int64 int64_t;
  typedef unsigned __int64 uint64_t;
# define INT64_C(x) x##I64
# define UINT64_C(x) x##UI64
# define PRId64 "I64d"
# define PRIu64 "I64u"
# define SCNd64 "I64d"
# define SCNu64 "I64u"
#else
# include <stdint.h>
# include <inttypes.h>
#endif

//--------------------------------------- C defines
# if __WORDSIZE == 64
#  define __PRI64_PREFIX	"l"
#  define __PRIPTR_PREFIX	"l"
# else
#  define __PRI64_PREFIX	"ll"
#  define __PRIPTR_PREFIX
# endif

/* Macros for printing format specifiers.  */

/* Decimal notation.  */
# define PRId8		"d"
# define PRId16		"d"
# define PRId32		"d"
# define PRId64		__PRI64_PREFIX "d"

/* Signed.  */
# define INT8_C(c)	c
# define INT16_C(c)	c
# define INT32_C(c)	c
# if __WORDSIZE == 64
#  define INT64_C(c)	c ## L
# else
#  define INT64_C(c)	c ## LL
# endif

/* Unsigned.  */
# define UINT8_C(c)	c
# define UINT16_C(c)	c
# define UINT32_C(c)	c ## U
# if __WORDSIZE == 64
#  define UINT64_C(c)	c ## UL
# else
#  define UINT64_C(c)	c ## ULL
# endif
//----------------------------------------

#if (defined(_M_IX86) && !defined(__i386__))
# define __i386__ 1
#endif
#if (defined(_M_X64) && !defined(__x86_64__))
# define __x86_64__ 1
#endif

/* Use the code from faster32.zip and if32.zip on 32-bit machines only. It
   works correctly in 64-bit mode too, but is slower.
*/
#if defined(__i386__) || defined(__ppc32__)
# ifndef FASTER32
#  define FASTER32 1
# endif
# ifndef IF32
#  define IF32 1
# endif
#endif

#ifdef BOINC
# include "boinc_api.h"
# include "filesys.h"
#endif

#ifndef EXIT_SUCCESS
# define EXIT_SUCCESS 0
#endif
#ifndef EXIT_FAILURE
# define EXIT_FAILURE 1
#endif


/***************************
 *
 * Start of compile options.
 *
 ***************************/

/* File names.
 */
#ifndef INI_FILENAME
# define INI_FILENAME "AP26-ini.txt"
#endif
#ifndef STATE_FILENAME
# define STATE_FILENAME "AP26-state.txt"
#endif
#ifndef RESULTS_FILENAME
# define RESULTS_FILENAME "SOL-AP26.txt"
#endif

/* Set MINIMUM_AP_LENGTH_TO_REPORT.
 */
#ifndef MINIMUM_AP_LENGTH_TO_REPORT
# define MINIMUM_AP_LENGTH_TO_REPORT 10
#endif

/* Set BUFFER_RESULTS=1 to keep the results file open between writes,
   flushing it to disk only during checkpoints.
*/
#ifndef BUFFER_RESULTS
# define BUFFER_RESULTS 1
#endif

/* Set REPORT_FREQUENCY to the number of times per K to report/checkpoint.
   The only valid values are 1, 7, 85, or 1355.
*/
#ifndef REPORT_FREQUENCY
# ifdef BOINC
#  define REPORT_FREQUENCY 85
# else
#  define REPORT_FREQUENCY 7
# endif
#endif

/*************************
 *
 * End of compile options.
 *
 *************************/


/* An alternative to GMP is provided for x86, x86_64 and ppx64. This is
   mainly to avoid the challenge of compiling the GMP library for Windows,
   but is also a little faster.
*/
#if defined(DUMMY)
# include "PrimeQ_dummy.h"
#elif ((defined(__i386__) || defined(__x86_64__)) && !defined(GMP))
# if defined(NO_LONG_DOUBLE)
#  include "PrimeQ_x86_nold.h"
# else
#  include "PrimeQ_x86.h"
# endif
#elif ((defined(__ppc64__) || defined(__powerpc64__)) && !defined(GMP))
# include "PrimeQ_ppc64.h"
#else
# include <gmp.h> 
# include "PrimeQ_gmp.h" 
# ifndef GMP
#  define GMP 1
# endif
#endif


static FILE *my_fopen(const char *filename, const char *mode)
{
#ifdef BOINC
  char resolved_name[512];

  boinc_resolve_filename(filename,resolved_name,sizeof(resolved_name));
  return boinc_fopen(resolved_name,mode);
#else
  return fopen(filename,mode);
#endif
}

static void write_state(int KMIN, int KMAX, int SHIFT, int K, int ITER)
{
  FILE *out;

  if ((out = my_fopen(STATE_FILENAME,"w")) == NULL)
  {
    fprintf(stderr,"Cannot open %s !!!\n",STATE_FILENAME);
    exit(EXIT_FAILURE);
  }

  if (fprintf(out,"%d %d %d %d %d %d\n",
              KMIN,KMAX,SHIFT,K,ITER,REPORT_FREQUENCY) < 0)
  {
    fprintf(stderr,"Cannot write to %s !!!\n",STATE_FILENAME);
    exit(EXIT_FAILURE);
  }

  fclose(out);
}

/* Return 1 only if a valid checkpoint can be read.
 */
static int read_state(int KMIN, int KMAX, int SHIFT, int *K, int *ITER)
{
  FILE *in;
  int tmp1, tmp2, tmp3, tmp6;

  if ((in = my_fopen(STATE_FILENAME,"r")) == NULL)
    return 0;

  if (fscanf(in,"%d %d %d %d %d %d\n",&tmp1,&tmp2,&tmp3,K,ITER,&tmp6) != 6)
  {
    fprintf(stderr,"Cannot parse %s !!!\n",STATE_FILENAME);
    exit(EXIT_FAILURE);
  }

  fclose(in);

  /* Check that KMIN KMAX SHIFT all match */
  if (tmp1==KMIN && tmp2==KMAX && tmp3==SHIFT)
  {
    if (tmp6!=REPORT_FREQUENCY)
    {
      /* We can still use K from this checkpoint, but must zero ITER. */
      *ITER = 0;
    }
    return 1;
  }

  return 0;
}


/* Global variables */
static int KMIN, KMAX, K_DONE, K_COUNT;

#if BUFFER_RESULTS
static FILE *results_file = NULL;
#endif


/* Returns index j where:
   0<=j<k ==> f+j*d*23# is composite.
   j=k    ==> for all 0<=j<k, f+j*d*23# is a strong probable prime to 9 bases.
*/
static int validate_ap26(int k, int d, int64_t f)
{
  int64_t N;
  int j;
#ifdef GMP
  mpz_t mpN;
#else
  const int base[] = {2,3,5,7,11,13,17,19,23};
#endif

  if (f%2==0)
    return 0;

#ifdef GMP
  mpz_init2(mpN,64);
#endif

  for (j = 0, N = f; j < k; j++)
  {
#if ((defined(__i386__) || defined(__x86_64__)) && !defined(GMP))
    int i;

    if (N < 0 || N >= (INT64_C(1) << LDBL_MULMOD_LIMIT))
    {
      fprintf(stderr,__FILE__": modulus out of range.\n");
      exit(EXIT_FAILURE);
    }

    for (i = 0; i < sizeof(base)/sizeof(int); i++)
      if (!strong_prp(base[i],N))
        return j;

#elif ((defined(__ppc64__) || defined(__powerpc64__)) && !defined(GMP))
    int i;

    if (N < 0)
    {
      fprintf(stderr,__FILE__": modulus out of range.\n");
      exit(EXIT_FAILURE);
    }

    for (i = 0; i < sizeof(base)/sizeof(int); i++)
      if (!strong_prp(base[i],N))
        return j;

#else
    if (N < 0)
    {
      fprintf(stderr,__FILE__": modulus out of range.\n");
      exit(EXIT_FAILURE);
    }

    if (sizeof(unsigned long) >= 8)
      mpz_set_ui(mpN,N);
    else
    {
      mpz_set_ui(mpN,N>>32);
      mpz_mul_2exp(mpN,mpN,32);
      mpz_add_ui(mpN,mpN,N&0xffffffffUL);
    }

    if (mpz_millerrabin(mpN,9) == 0)
      break;
#endif

    N += (int64_t)d*2*3*5*7*11*13*17*19*23;
  }

#ifdef GMP
  mpz_clear(mpN);
#endif

  return j;
}

// define what action to take when a solution is found
void ReportSolution(int AP_Length,int difference,int64_t First_Term)
{
  if (AP_Length >= MINIMUM_AP_LENGTH_TO_REPORT)
  {
    int i;
#if !BUFFER_RESULTS
    FILE *results_file = NULL;
#endif

    if ((i = validate_ap26(AP_Length,difference,First_Term)) < AP_Length)
    {
#ifdef BOINC
      if (boinc_is_standalone())
#endif
        printf("Non-Solution: %d %d %"PRId64"\n",
               AP_Length,difference,First_Term);  

      /* Check leading terms */
      ReportSolution(i,difference,First_Term);

      /* Check trailing terms */
      ReportSolution(AP_Length-(i+1),difference,First_Term
                     +(int64_t)(i+1)*difference*2*3*5*7*11*13*17*19*23);
      return;
    }

    if (results_file == NULL)
      results_file = my_fopen(RESULTS_FILENAME,"a");

#ifdef BOINC
      if (boinc_is_standalone())
#endif
        printf("Solution: %d %d %"PRId64"\n",
               AP_Length,difference,First_Term);

    if (results_file == NULL)
    {
      fprintf(stderr,"Cannot open %s !!!\n",RESULTS_FILENAME);
      exit(EXIT_FAILURE);
    }

    if (fprintf(results_file,"%d %d %"PRId64"\n",
                AP_Length,difference,First_Term)<0)
    {
      fprintf(stderr,"Cannot write to %s !!!\n",RESULTS_FILENAME);
      exit(EXIT_FAILURE);
    }

#if !BUFFER_RESULTS
    fclose(results_file);
#endif
  }
}


/* Checkpoint at iteration ITER, where 0 <= ITER < REPORT_FREQUENCY.
   If force is nonzero then don't ask BOINC for permission.
*/
void checkpoint(int SHIFT, int K, int ITER, int force)
{
  double d;

  if (K_COUNT > 0)
    d = (double)(K_DONE*REPORT_FREQUENCY+ITER)/(K_COUNT*REPORT_FREQUENCY);
  else
    d = 1.0;

#ifdef BOINC
  if (force)
    boinc_begin_critical_section();
  if (force || boinc_time_to_checkpoint())
  {
#endif

#if BUFFER_RESULTS
    if (results_file != NULL)
      fflush(results_file);
#endif
    write_state(KMIN,KMAX,SHIFT,K,ITER);

    /* It is OK to print this progress message now since other buffers are
       being flushed now anyway */
    printf("Checkpoint: KMIN=%d KMAX=%d SHIFT=%d K=%d ITER=%d/%d (%.2f%%)\n",
           KMIN,KMAX,SHIFT,K,ITER,REPORT_FREQUENCY,d*100.0);

#ifdef BOINC
    if (!force)
      boinc_checkpoint_completed();
  }
  if (force)
    boinc_end_critical_section();
#endif

#ifdef BOINC
  boinc_fraction_done(d);
#endif
}


#include "AP26-boinc.h"
#include "AP26-boinc-SSE.h"
#include "AP26-boinc-SSE2.h"


/* Returns 1 iff K will be searched.
 */
static int will_search(int K)
{
  return (K%PRIME1 && K%PRIME2 && K%PRIME3 && K%PRIME4 &&
          K%PRIME5 && K%PRIME6 && K%PRIME7 && K%PRIME8);
}


int main(int argc, char *argv[])
{
  int i, K, SHIFT, ITER, LINEAR = 0;

  
  printf("Compiled " __DATE__ " with Intel(R) C++ Compiler 11.1 (32-bit)\n");

  // bryan little, detect mmx/sse2 or use generic code path
    char vendor[13];        // 12 bytes + null char
    bool sseFlag;           // edx:25
    bool sse2Flag;          // edx:26

    unsigned int maxCalls;
    unsigned int regEax, regEdx;

    vendor[0] = '\0';
    sseFlag = sse2Flag = false;

    // MSVC specific inline assembly block ////////////////////////////////////

    // check cpuid is supported
    // If bit 21 (22nd bit) in EFLAGS can be toggled, cpuid is supported
    __asm {
        pushfd
        pop     eax                 // get flags
        mov     ebx,    eax         // copy the flags

        xor     eax,    200000h     // reverse bit 21 (1->0, or 0->1)
        push    eax
        popfd                       // set flags
        pushfd
        pop     eax                 // re-get flags

        xor     eax,    ebx         // compare previous flags
        mov     regEax, eax         // it returns 0x200000, if bit 21 can be toggled, otherwise returns 0
    }

    // CPU does not support cpuid instruction, stop here
	if(regEax){

		__asm {
			// first need to set eax to 0
			xor     eax,    eax

			// when call cpuid with eax=0, cpuid returns max number of cpuid calls and vendor string (ebx-edx-ecx)
			cpuid

			// copy max cpuid calls to maxCalls
			mov     maxCalls,   eax

			// copy vendor string (12 bytes, ebx-ecx-edx)
			mov     dword ptr [vendor],     ebx
			mov     dword ptr [vendor+4],   edx
			mov     dword ptr [vendor+8],   ecx
		}

		vendor[12] = 0; // append 0

		// if maxCalls is greater than 1, check supported features
		if(maxCalls < 1)
		{
			sseFlag = sse2Flag = false;
		}
		else{

			__asm {
				// call cpuid once again with eax=1
				mov     eax,      01h
				cpuid

				// get results into local vars
				mov     regEax,   eax
				mov     regEdx,   edx
			}

			sseFlag = (regEdx >> 25) & 0x1;

			sse2Flag = (regEdx >> 26) & 0x1;

		}

		printf("Processor: %s", vendor);
		if(sse2Flag){
			printf(" supporting SSE2\n");
		}
		else if(sseFlag){
			printf(" supporting SSE\n");
		}

	}

// end cpu detection



#ifdef DUMMY
  printf("WARNING: This is a dummy executable for testing purposes only.\n");
#endif

#ifdef BOINC
  boinc_init();
#endif

#if ((defined(__i386__) || defined(__x86_64__)) && !defined(GMP))
  /* Check FPU mode and change it to extended precision if necessary. */
  check_fpu_mode();
#endif

  /* Get search parameters from command line or ini file */
  if (argc == 1)
  {
    FILE *in;

    if ((in = my_fopen(INI_FILENAME,"r")) == NULL)
    {
      fprintf(stderr,"Cannot open %s !!!\n",INI_FILENAME);
      exit(EXIT_FAILURE);
    }

    if (fscanf(in,"%d %d %d %d\n",&KMIN,&KMAX,&SHIFT,&LINEAR) < 3)
    {
      fprintf(stderr,"Cannot parse %s !!!\n",INI_FILENAME);
      exit(EXIT_FAILURE);
    }

    fclose(in);
  }
  else if (argc == 4 || argc == 5)
  {
    sscanf(argv[1],"%d",&KMIN);
    sscanf(argv[2],"%d",&KMAX);
    sscanf(argv[3],"%d",&SHIFT);
    if (argc == 5)
      sscanf(argv[4],"%d",&LINEAR);
  }
  else
  {
    fprintf(stderr,"Usage: %s [KMIN KMAX SHIFT [LINEAR]]\n",argv[0]);
    exit(EXIT_FAILURE);
  }

  if (LINEAR && 0 < KMIN && KMIN <= KMAX)
  {
    int j, k, n0 = KMIN, n1 = KMAX;
    int w29=29, w31=31, w37=37, w41=41, w43=43, w47=47, w53=53, w59=59;

    for (i = 0, k = 0; i < n0; k++)
    {
      j = 1;
      if (--w29 == 0) w29 = 29, j = 0;
      if (--w31 == 0) w31 = 31, j = 0;
      if (--w37 == 0) w37 = 37, j = 0;
      if (--w41 == 0) w41 = 41, j = 0;
      if (--w43 == 0) w43 = 43, j = 0;
      if (--w47 == 0) w47 = 47, j = 0;
      if (--w53 == 0) w53 = 53, j = 0;
      if (--w59 == 0) w59 = 59, j = 0;
      i += j;
    }
    KMIN = k;

    for (i = n0; i < n1; k++)
    {
      j = 1;
      if (--w29 == 0) w29 = 29, j = 0;
      if (--w31 == 0) w31 = 31, j = 0;
      if (--w37 == 0) w37 = 37, j = 0;
      if (--w41 == 0) w41 = 41, j = 0;
      if (--w43 == 0) w43 = 43, j = 0;
      if (--w47 == 0) w47 = 47, j = 0;
      if (--w53 == 0) w53 = 53, j = 0;
      if (--w59 == 0) w59 = 59, j = 0;
      i += j;
    }
    KMAX = k;

    printf("Linear parameters transform to: KMIN=%d, KMAX=%d\n",KMIN,KMAX);
  }

  /* Resume from checkpoint if there is one */
  if (read_state(KMIN,KMAX,SHIFT,&K,&ITER))
  {
    printf("Resuming search from the checkpoint in %s.\n",STATE_FILENAME);
  }
  else
  {
    printf("Beginning a new search with parameters from %s.\n",
           (argc == 1)? INI_FILENAME : "the command line");
    K = KMIN;
    ITER = 0;
  }

  /* Count the number of K in the range KMIN <= K <= KMAX that will actually
     be searched and (if K > KMIN) those that have already been searched. */
  for (i = KMIN; i <= KMAX; i++)
    if (will_search(i))
    {
      K_COUNT++;
      if (K > i)
        K_DONE++;
    }

  printf("%d K in this range remaining to be searched (%d skipped, %d done)."
         "\n",K_COUNT-K_DONE,KMAX+1-KMIN-K_COUNT,K_DONE);

  /* Top-level loop */
  for (; K <= KMAX; ++K)
    if (will_search(K))
    {
#if (REPORT_FREQUENCY > 1)
		if(sse2Flag){
			printf("Starting SSE2 optimized code\n");
			SearchAP26_SSE2(K,SHIFT,ITER);
		}
		else if(sseFlag){
			printf("Starting SSE optimized code\n");
			SearchAP26_SSE(K,SHIFT,ITER);
		}
		else{ //generic IA32 code
			printf("Starting generic C code\n");
			SearchAP26(K,SHIFT,ITER);
		}
//		SearchAP26(K,SHIFT,ITER);
		ITER = 0;
#else
		 checkpoint(SHIFT,K,0,0);
		 SearchAP26(K,SHIFT);
#endif
		 K_DONE++;
    }

  /* Force Final checkpoint */
  checkpoint(SHIFT,K,0,1);

  /* Done */
#ifdef BOINC
  boinc_finish(0);
#endif

#ifdef DUMMY
  printf("PrimeQ() was called %"PRIu64" times.\n",PrimeQ_call_count);
#endif

  return EXIT_SUCCESS;
}
